/*
 * Copyright 1997-2013 Optimatika (www.optimatika.se)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package org.ojalgo.concurrent;

import java.util.concurrent.ForkJoinTask;

import org.ojalgo.OjAlgoUtils;
import org.ojalgo.type.IntCount;

/**
 * @author apete
 */
public abstract class DivideAndConquer extends Object {

    public static interface Strategy {

        IntCount getInitialWorkerCount();

        boolean shouldDivideFurther(int localIndexCount, IntCount localWorkerCount);

    }

    static class DefaultStrategy implements Strategy {

        private final int myIndexCountThreshold;

        DefaultStrategy(final int initialIndexCount, final int externalThreshold) {

            super();

            myIndexCountThreshold = Math.max(1, (externalThreshold * externalThreshold) / initialIndexCount);
        }

        public IntCount getInitialWorkerCount() {
            return OjAlgoUtils.ENVIRONMENT.countThreads();
        }

        public boolean shouldDivideFurther(final int localIndexCount, final IntCount localWorkerCount) {
            return ConcurrentUtils.shouldBranch(localIndexCount, myIndexCountThreshold, localWorkerCount);
        }

    }

    private final int myThreshold;

    public DivideAndConquer(final int aThreshold) {

        super();

        myThreshold = aThreshold;
    }

    @SuppressWarnings("unused")
    private DivideAndConquer() {
        this(2);
    }

    /**
     * Asynchronous execution - start it, and forget about it.
     * 
     * @param aFirst The first index, in a range, to include.
     * @param aLimit The first index NOT to include - last (excl.) index in a range.
     */
    public final void execute(final int aFirst, final int aLimit) {

        final int tmpCount = ConcurrentUtils.calculateIndexCount(aFirst, aLimit);

        final int tmpThreshold = Math.max(1, (myThreshold * myThreshold) / tmpCount);

        DaemonPoolExecutor.INSTANCE.execute(new ForkJoinTask<Void>() {

            @Override
            public Void getRawResult() {
                return null;
            }

            @Override
            protected boolean exec() {
                DivideAndConquer.this.divide(aFirst, aLimit, tmpThreshold, OjAlgoUtils.ENVIRONMENT.countThreads());
                return true;
            }

            @Override
            protected void setRawResult(final Void value) {
            }
        });
    }

    /**
     * Synchronous execution - wait until it's finished.
     * 
     * @param aFirst The first index, in a range, to include.
     * @param aLimit The first index NOT to include - last (excl.) index in a range.
     */
    public final void invoke(final int aFirst, final int aLimit) {

        final int tmpCount = ConcurrentUtils.calculateIndexCount(aFirst, aLimit);

        final int tmpThreshold = Math.max(1, (myThreshold * myThreshold) / tmpCount);

        DaemonPoolExecutor.INSTANCE.invoke(new ForkJoinTask<Void>() {

            @Override
            public Void getRawResult() {
                return null;
            }

            @Override
            protected boolean exec() {
                DivideAndConquer.this.divide(aFirst, aLimit, tmpThreshold, OjAlgoUtils.ENVIRONMENT.countThreads());
                return true;
            }

            @Override
            protected void setRawResult(final Void value) {
            }
        });
    }

    protected abstract void conquer(final int aFirst, final int aLimit);

    protected final void divide(final int aFirst, final int aLimit, final int indexCountThreshold, final IntCount locallyAvailableWorkers) {

        final int tmpIndexCount = ConcurrentUtils.calculateIndexCount(aFirst, aLimit);

        if (ConcurrentUtils.shouldBranch(tmpIndexCount, indexCountThreshold, locallyAvailableWorkers)) {

            final int tmpSplit = ConcurrentUtils.calculateSplitIndex(aFirst, tmpIndexCount);
            final IntCount tmpWorkers = locallyAvailableWorkers.halve();

            final ForkJoinTask<Void> tmpForkedTask = new ForkJoinTask<Void>() {

                @Override
                public Void getRawResult() {
                    return null;
                }

                @Override
                protected boolean exec() {
                    DivideAndConquer.this.divide(aFirst, tmpSplit, indexCountThreshold, tmpWorkers);
                    return true;
                }

                @Override
                protected void setRawResult(final Void value) {
                }

            };

            tmpForkedTask.fork();

            DivideAndConquer.this.divide(tmpSplit, aLimit, indexCountThreshold, tmpWorkers);

            tmpForkedTask.join();

        } else {

            this.conquer(aFirst, aLimit);
        }
    }

}
